import * as tslib_1 from "tslib";
import Header from "./header";
import { autobind } from "core-decorators";
import * as React from "react";
import DataRowComponent, { VirtualizedDataRow } from "./rows";
const lazyload = require("react-lazyload");
export function isRootColumn(col) {
    return "subColumns" in col && !!col.subColumns;
}
export function rootColumnIndex(columns, i) {
    const cols = columns.slice(0, i + 1);
    return cols.reduce((count, col) => {
        if (isRootColumn(col)) {
            return col.subColumns.length + count;
        }
        else {
            return count + 1;
        }
    }, 0) - 1;
}
export function subColumnIndex(columns, i, rootIndex) {
    const cols = columns.slice(0, rootIndex);
    return cols.reduce((count, col) => {
        if (isRootColumn(col)) {
            return col.subColumns.length + count;
        }
        else {
            return count + 1;
        }
    }, 0) + i;
}
let externalStylesheet;
const isModule = (s) => Object.keys(s).length > 0;
export function loadExternalStyleSheet(extSheet) {
    externalStylesheet = extSheet;
    styleSheetParser = loadStyle();
}
function loadStyle() {
    try {
        const stylesheet = require("./index.scss");
        const s = new Proxy({}, {
            get: (_, name) => [stylesheet, externalStylesheet]
                .filter(s => s)
                .map(s => isModule(s) ? stylesheet[name] : name)
                .join(" "),
        });
        return s;
    }
    catch (_) {
        return {};
    }
}
export let styleSheetParser = loadStyle();
function isNullOrUndefined(v) {
    return v === null || typeof v === "undefined";
}
export function isCategorizedRow(row) {
    return "category" in row && !!row.category;
}
export function areCategorizedRows(rows) {
    return rows.every((row) => isCategorizedRow(row));
}
function getCategorizedRows(rows) {
    const categorizedRows = {};
    if (areCategorizedRows(rows)) {
        rows.forEach((row) => categorizedRows[row.category] = (categorizedRows[row.category] || []).concat(row));
    }
    else {
        categorizedRows[""] = rows;
    }
    return categorizedRows;
}
let CubeDataTable = class CubeDataTable extends React.PureComponent {
    constructor(props, context) {
        super(props, context);
        this.scrollableParent = null;
        const sortableColumns = props.columns.filter((col) => !isRootColumn(col));
        this.state = {
            ascendant: false,
            currentSortedColumn: sortableColumns.findIndex((col) => col.featured && !!col.sort || false),
            currentSorting: (sortableColumns.find((col) => col.featured && !!col.sort || false) || { sort: undefined }).sort,
        };
        this.validateRowsLength(props);
    }
    onRef(tableDom) {
        this.tableDom = tableDom;
        this.setScrollableParent(this.getScrollableParent(this.tableDom && this.tableDom.parentElement));
    }
    getScrollableParent(parentStack) {
        while (parentStack) {
            if (parentStack.scrollHeight > parentStack.clientHeight
                && ["scroll", "auto"].indexOf(window.getComputedStyle(parentStack).overflowY) !== -1) {
                return parentStack;
            }
            parentStack = parentStack.parentElement;
        }
        return null;
    }
    setScrollableParent(scrollable) {
        if (this.scrollableParent !== scrollable) {
            this.removeScrollListeners();
            if (scrollable)
                scrollable.addEventListener("scroll", this.scrollHandleSticky, { capture: true, passive: true });
        }
        this.scrollableParent = scrollable;
    }
    setStickyTablePosition() {
        if (this.scrollableParent && this.stickyTable && this.stickyTable.parentElement) {
            const tableRect = this.tableDom.getBoundingClientRect();
            const tableTop = tableRect.top;
            const tableHeight = tableRect.height;
            const parentTop = this.scrollableParent.getBoundingClientRect().top;
            const stickyHeight = this.stickyTable.getBoundingClientRect().height;
            const pixelsToEnd = tableTop + tableHeight - parentTop;
            this.stickyTable.style.top = (parentTop - tableTop + Math.min(0, pixelsToEnd - stickyHeight)) + "px";
        }
    }
    windowHandleSticky() {
        return this.handleSticky(true);
    }
    scrollHandleSticky() {
        return this.handleSticky(false);
    }
    handleSticky(forceRerender) {
        lazyload.forceCheck();
        if (this.tableDom && this.scrollableParent) {
            const tableRect = this.tableDom.getBoundingClientRect();
            const parentRectTop = this.scrollableParent.getBoundingClientRect().top;
            const isSticky = tableRect.top < parentRectTop;
            const pixelsToEnd = tableRect.top + tableRect.height - parentRectTop;
            const stickyRuleRow = this.getCurrentStickyRuleRow();
            if (stickyRuleRow !== this.state.stickyRuleRow) {
                this.setState({ stickyRuleRow });
            }
            if (isSticky && this.stickyTable) {
                this.setStickyTablePosition();
            }
            if (this.state.isSticky !== isSticky && pixelsToEnd > 0) {
                this.setState({ isSticky });
            }
            else if (pixelsToEnd <= 0 && this.state.isSticky !== false) {
                this.setState({ isSticky: false });
            }
            if (forceRerender) {
                this.syncStickyTableStyle();
            }
        }
    }
    syncStickyTableStyle() {
        if (this.stickyTable && this.scrollableParent && this.tableDom) {
            const originalHeader = this.tableDom.querySelector("thead");
            const boundingRect = originalHeader.getBoundingClientRect();
            this.stickyTable.parentElement.style.width = boundingRect.width + "px";
            const originalRows = Array.from(originalHeader.querySelectorAll("tr"));
            const stickyRows = Array.from(this.stickyTable.querySelectorAll("tr"));
            for (const rowIndex in originalRows) {
                const originalRow = originalRows[rowIndex];
                const stickyRow = stickyRows[rowIndex];
                const originalCells = Array.from(originalRow.querySelectorAll("td, th"));
                const stickyCells = Array.from(stickyRow.querySelectorAll("td, th"));
                for (const cellIndex in originalCells) {
                    const stickyCell = stickyCells[cellIndex];
                    const originalCell = originalCells[cellIndex];
                    stickyCell.style.width = originalCell.getBoundingClientRect().width + "px";
                }
            }
        }
    }
    componentWillReceiveProps(nextProps) {
        if (this.props.data === nextProps.data && this.props.columns === nextProps.columns)
            return;
        this.validateRowsLength(nextProps);
    }
    validateRowsLength(props) {
        if (process && process.env && process.env.NODE_ENV === "production")
            return;
        if (!props.columns || (!props.data && !props.featuredRow))
            return;
        const rows = props.featuredRow ? [...(props.data || []), props.featuredRow] : (props.data || []);
        const columns = props.columns.reduce((count, col) => count + (isRootColumn(col) ? col.subColumns.length : 1), 0);
        const invalidRows = rows.filter(row => row.data.length !== columns);
        if (invalidRows.length > 0) {
            console.groupCollapsed("Warning: invalid CubeDataTable props");
            console.warn("Header has " + columns + " columns and not all rows have the same cell count");
            for (const row of invalidRows) {
                console.warn(`Row #${rows.findIndex(r => r.id === row.id)} has invalid cell count (${row.data.length}): `, row);
            }
            console.groupEnd();
        }
    }
    generateRuleRow(content) {
        const rows = this.props.data.filter(r => r.category === content);
        const nOfCols = rootColumnIndex(this.props.columns, this.props.columns.length - 1) + 1;
        const noCategory = "Sem categoria";
        return (React.createElement("tr", { key: content || noCategory, className: styleSheetParser.ruleRow, "data-row-type": "rule" },
            this.props.onSelect && !this.props.hideSelects ? (React.createElement("th", { className: styleSheetParser.selectHeader },
                React.createElement("input", { type: "checkbox", onChange: () => this.toggleAllGivenRows(rows), checked: rows.length && rows.every(r => !!r.selected) || false }))) : null,
            React.createElement("td", { colSpan: nOfCols }, content || noCategory)));
    }
    renderRow(row, index, isTotal) {
        const siblings = Math.max(...row.data.map(cell => cell && (cell.node ? (Array.isArray(cell.node)) : Array.isArray(cell.value)) ? (cell.node || cell.value).length : 1));
        const DataRowComp = this.props.virtualize ? VirtualizedDataRow : DataRowComponent;
        return [...new Array(siblings)].map((_, i) => React.createElement(DataRowComp, Object.assign({}, row, { sibling: i, index: index, hideSelect: isTotal, key: `${row.id}@${i}`, onSelect: !this.props.hideSelects ? this.props.onSelect : undefined, className: `${isTotal ? styleSheetParser.featuredRow : ""} ${row.className}` })));
    }
    generateCategorizedRows(rows) {
        const elements = [];
        const categorized = getCategorizedRows(rows);
        const categories = Object.entries(categorized).sort(([a], [b]) => a.localeCompare(b));
        for (const [category, categoryRows] of categories) {
            if (category) {
                elements.push(this.generateRuleRow(category));
            }
            if (categoryRows) {
                elements.push(...categoryRows
                    .map((row, i) => this.renderRow(row, i))
                    .reduce((acc, arr) => acc.concat(arr), []));
                if (this.props.totalFunction) {
                    elements.push(...this.renderRow(this.props.totalFunction(categoryRows, category), 0, true));
                }
            }
        }
        return elements;
    }
    sortData(data) {
        return data.sort((r1, r2) => {
            const currentSorting = this.state.currentSorting;
            const currentSortedColumn = this.state.currentSortedColumn;
            if (!isNullOrUndefined(currentSorting) && !isNullOrUndefined(currentSortedColumn)) {
                if (!r1.data[currentSortedColumn] && r2.data[currentSortedColumn]) {
                    return -1;
                }
                else if (!r2.data[currentSortedColumn] && r1.data[currentSortedColumn]) {
                    return 1;
                }
                else if (!r1.data[currentSortedColumn] && !r2.data[currentSortedColumn]) {
                    return 0;
                }
                return currentSorting(r1.data[currentSortedColumn].value, r2.data[currentSortedColumn].value);
            }
            else {
                return 0;
            }
        });
    }
    onSortRootColumn(colId) {
        const col = this.props.columns[colId];
        this.setState({
            currentSorting: col.sort,
            currentSortedColumn: rootColumnIndex(this.props.columns, colId),
            ascendant: (col.sort === this.state.currentSorting) ? !this.state.ascendant : false
        });
    }
    onSort(rootColIndex, subColIndex) {
        if (typeof subColIndex !== "undefined") {
            this.onSortSubColumn(rootColIndex, subColIndex);
        }
        else {
            this.onSortRootColumn(rootColIndex);
        }
    }
    onSortSubColumn(rootColId, subColId) {
        const subCol = this.props.columns[rootColId].subColumns[subColId];
        this.setState({
            ascendant: (subCol.sort === this.state.currentSorting) ? !this.state.ascendant : false,
            currentSortedColumn: subColumnIndex(this.props.columns, subColId, rootColId),
            currentSorting: subCol.sort,
        });
    }
    render() {
        if (this.props.placeholder && (!this.props.data || this.props.data.length === 0)) {
            return React.createElement("div", null, this.props.placeholder);
        }
        return this.renderTable();
    }
    componentDidMount() {
        window.addEventListener("resize", this.windowHandleSticky);
    }
    componentWillUnmount() {
        window.removeEventListener("resize", this.windowHandleSticky);
        this.removeScrollListeners();
    }
    removeScrollListeners() {
        if (this.scrollableParent) {
            this.scrollableParent.removeEventListener("scroll", this.scrollHandleSticky);
        }
    }
    componentDidUpdate() {
        if (!this.scrollableParent) {
            this.onRef(this.tableDom);
        }
    }
    onStickyTableRef(stickyHeader) {
        if (this.stickyTable !== stickyHeader) {
            this.stickyTable = stickyHeader;
            this.setStickyTablePosition();
            this.syncStickyTableStyle();
        }
    }
    genHeader(sticky) {
        const header = this.getHeaderElement();
        if (sticky) {
            return (React.createElement("table", { className: `${styleSheetParser.portal} ${styleSheetParser.table}`, ref: sticky && this.onStickyTableRef || undefined },
                header,
                React.createElement("tbody", null, this.state.stickyRuleRow && this.generateRuleRow(this.state.stickyRuleRow || ""))));
        }
        else {
            return header;
        }
    }
    getCurrentStickyRuleRow() {
        const headerBottom = this.stickyTable && this.stickyTable.getBoundingClientRect().bottom || 0;
        const ruleRows = Array.from(this.tableDom.querySelectorAll(`tr[data-row-type="rule"]`));
        const scrolledRuleRows = ruleRows.filter(row => row.getBoundingClientRect().top <= headerBottom);
        const lastScrolledRuleRow = scrolledRuleRows.reduce((r1, r2) => {
            if (!r1)
                return r2;
            const t1 = r1.getBoundingClientRect().top;
            const t2 = r2.getBoundingClientRect().top;
            return t1 > t2 ? r1 : r2;
        }, null);
        return lastScrolledRuleRow && lastScrolledRuleRow.textContent || undefined;
    }
    areAllRowsSelected(rows = this.props.data) {
        return rows.every((row) => !!row.selected);
    }
    toggleAllRows() {
        return this.toggleAllGivenRows();
    }
    toggleAllGivenRows(rs) {
        if (!this.props.onSelect) {
            console.warn("CubeDataTable: invalid call to toggleAllRows");
            return;
        }
        const rows = rs || this.props.data;
        this.props.onSelect(rows.map(row => row.id), !this.areAllRowsSelected(rs));
    }
    getHeaderElement() {
        return (React.createElement(Header, { areAllRowsSelected: this.areAllRowsSelected(), currentSortedColumn: this.state.currentSortedColumn, onToggleAll: this.toggleAllRows, hideSelect: !this.props.onSelect || this.props.hideSelects || false, columns: this.props.columns, ascendant: this.state.ascendant, headerCellClassName: this.props.headerCellClassName, onSort: this.onSort }));
    }
    renderTable() {
        let sortedData = this.state.currentSorting ? this.sortData(this.props.data) : this.props.data;
        if (this.state.currentSorting && this.state.ascendant) {
            sortedData = sortedData.reverse();
        }
        return (React.createElement("div", null,
            React.createElement("div", { className: styleSheetParser.wrapper },
                this.state.isSticky ? (this.genHeader(true)) : null,
                React.createElement("table", { ref: this.onRef, className: styleSheetParser.table + " " + (this.props.className || "") + " " + styleSheetParser.tableResponsive },
                    this.genHeader(),
                    React.createElement("tbody", null,
                        this.props.featuredRow ? (this.renderRow(this.props.featuredRow, 0, true)) : null,
                        this.generateCategorizedRows(sortedData))))));
    }
};
CubeDataTable = tslib_1.__decorate([
    autobind
], CubeDataTable);
export default CubeDataTable;
//# sourceMappingURL=index.js.map